#include <libusb.h>
#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>
#include <dlfcn.h>
#include <string.h>
#include <linux/limits.h>
#include <unistd.h>
#include <libgen.h>

#include "device/device.h"
#include "utils/hashmap.h"

hashmap_t* usb_drivers;

void usb_drivers_init()
{
    // Initialize hashmap
    usb_drivers = hashmap_alloc();

    // Obtain current executable path
    char exe_path[PATH_MAX];
    readlink("/proc/self/exe", exe_path, PATH_MAX);
    
    char* path = dirname(exe_path);
    size_t path_len = strlen(path);

    char driver_path[PATH_MAX];
    memcpy(driver_path, path, path_len);
    memcpy(driver_path + path_len, "/drivers/", 10);

    // Open driver directory and load all symbols
    DIR* directory = opendir(driver_path);
    if(!directory)
    {
        fprintf(stderr, "Error: Could not open '%s' directory\n", driver_path);
        return;
    }

    printf("Opened drivers directory: %s\n", driver_path);

    struct dirent* entry;
    while((entry = readdir(directory)))
    {
        if(entry->d_type & DT_DIR) continue;

        size_t entry_name_len = strlen(entry->d_name);

        char fullpath[PATH_MAX];
        memcpy(fullpath, driver_path, path_len + 9);
        memcpy(fullpath + path_len + 9, entry->d_name, entry_name_len);
        *(fullpath + path_len + 9 + entry_name_len) = 0;
        
        void* driver = dlopen(fullpath, RTLD_LAZY | RTLD_LOCAL);
        void (*init_fn)(void) = dlsym(driver, "driver_init");
        init_fn();

        uint32_t (*getkey_fn)(void) = dlsym(driver, "driver_getkey");
        uint32_t key;
        while((key = getkey_fn()))
            hashmap_put(usb_drivers, key, driver);
    }

    closedir(directory);
}

void usb_init()
{
    // Init drivers
    usb_drivers_init();

    // Init libusb
    if(libusb_init(NULL))
    {
        fprintf(stderr, "Error: Could not initialize libusb\n");
        return;
    }

    // List all usb devices obtaining 'vendor:product'
    // Call corresponding driver if it exists, so 
    // that we get a peripheral handle
    libusb_device** list;
    ssize_t device_count = libusb_get_device_list(NULL, &list);
    if(device_count <= 0) return;

    for(ssize_t i = 0; i < device_count; i++)
    {
        libusb_device* current = list[i];
        struct libusb_device_descriptor desc;
        libusb_get_device_descriptor(current, &desc);

        uint32_t key = (desc.idVendor << 16) | desc.idProduct;
        void* driver = hashmap_get(usb_drivers, key);
        if(!driver) continue;

        printf("Found driver for usb device %04x:%04x\n", desc.idVendor, desc.idProduct);

        device_register(driver, current);
    }
}
